const { app, BrowserWindow, Menu } = require("electron");
const path = require("path");
const express = require("express");
const { createBareServer } = require("bare-server-node"); // Import bare server
const http = require("http"); // Import HTTP module for bare server

// Express server setup
const server = express();

// Force English language
server.use((req, res, next) => {
  req.headers['accept-language'] = 'en-US,en;q=0.9';
  next();
});

// Serve static files from the "public" folder
server.use(express.static(path.join(__dirname, "public")));

// Serve the UV-Static files from the "active" folder
server.use("/active", express.static(path.join(__dirname, "public/active")));

const PORT = 3000;
server.listen(PORT, () => {
  console.log(`Express server running at http://localhost:${PORT}`);
});

// Bare server setup
const bareServer = createBareServer('/bare/'); // Create bare server instance
const barePort = 4000; // Use a different port for the bare server

// Create an HTTP server for the bare server
const bareHttpServer = http.createServer();

// Handle incoming requests for the bare server
bareHttpServer.on('request', (req, res) => {
  if (bareServer.shouldRoute(req)) {
    bareServer.routeRequest(req, res);
  } else {
    res.writeHead(400);
    res.end('Not a bare request');
  }
});

// Handle WebSocket upgrades for the bare server
bareHttpServer.on('upgrade', (req, socket, head) => {
  if (bareServer.shouldRoute(req)) {
    bareServer.routeUpgrade(req, socket, head);
  } else {
    socket.end();
  }
});

// Start the bare server
bareHttpServer.listen(barePort, () => {
  console.log(`Bare server running at http://localhost:${barePort}`);
});

let mainWindow;

// Custom menu template
const template = [
  {
    label: "Help",
    submenu: [
      {
        label: "How to Use",
        click: () => {
          mainWindow.loadURL(`http://localhost:${PORT}/readme.html`);
        },
      },
    ],
  },
  {
    label: "Window",
    submenu: [
      { role: "minimize" }, // Minimize window
      { role: "zoom" }, // Zoom
      { type: "separator" }, // Separator
      { role: "close" }, // Close window
    ],
  },
  {
    label: "View",
    submenu: [
      { role: "zoomIn" }, // Zoom in
      { role: "zoomOut" }, // Zoom out
      { role: "togglefullscreen" }, // Full screen
    ],
  },
  {
    label: "Restart",
    click: () => {
      app.relaunch(); // Relaunch the app
      app.exit(); // Exit the current instance
    },
  },
{
    label: "Refresh",
    click: () => {
      if (mainWindow) {
        mainWindow.reload(); // Refresh the page
      }
    },
  },
];

// Create the menu from the template
const menu = Menu.buildFromTemplate(template);
Menu.setApplicationMenu(menu);

app.whenReady().then(() => {
  mainWindow = new BrowserWindow({
    width: 800,
    height: 600,
    webPreferences: {
      nodeIntegration: false, // Keep security in mind
      contextIsolation: true,
    },
  });


  // Load the main index.html file
  mainWindow.loadURL(`http://localhost:${PORT}/index.html`);
});